@@ -1,5 +1,37 @@
 /*
  * HTTP redirection functions.
+ *
+ * Copyright (c) 2006, Jens Jakobsen 
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ *
+ *   Redistributions of source code must retain the above copyright notice,
+ *   this list of conditions and the following disclaimer.
+ *
+ *   Redistributions in binary form must reproduce the above copyright
+ *   notice, this list of conditions and the following disclaimer in the
+ *   documentation and/or other materials provided with the distribution.
+ *
+ *   Neither the names of copyright holders nor the names of its contributors
+ *   may be used to endorse or promote products derived from this
+ *   software without specific prior written permission.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+ * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
+ * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ *
+ *
  * Copyright (C) 2004, 2005 Mondru AB.
  *
  * The contents of this file may be used under the terms of the GNU
@@ -9,6 +41,7 @@
  *
  */
 
+#include <sys/types.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
 
@@ -214,8 +247,9 @@ static int redir_stradd(char *dst, int dstsize, char *fmt, ...) {
 
 
 /* Make an XML Reply */
-static int redir_xmlreply(struct redir_t *redir, int res, char* hexchal, 
-			  char* reply, char* redirurl, 
+static int redir_xmlreply(struct redir_t *redir, struct redir_conn_t *conn, 
+			  int res, long int timeleft, char* hexchal, 
+			  char* reply, char* redirurl,
 			  char *dst, int dstsize) {
   char mid[REDIR_MAXBUFFER];
 
@@ -323,6 +357,56 @@ static int redir_xmlreply(struct redir_t *redir, int res, char* hexchal,
 		 inet_ntoa(redir->addr), redir->port);
     redir_stradd(dst, dstsize, "</AbortLoginReply>\r\n");
     break;
+  case REDIR_STATUS:
+    redir_stradd(dst, dstsize, "<AuthenticationPollReply>\r\n");
+    redir_stradd(dst, dstsize, "<MessageType>140</MessageType>\r\n");
+    if (conn->authenticated != 1) {
+      redir_stradd(dst, dstsize, "<ResponseCode>150</ResponseCode>\r\n");
+      redir_stradd(dst, dstsize, 
+		   "<ReplyMessage>Not logged on</ReplyMessage>\r\n");
+    }
+    else {
+      redir_stradd(dst, dstsize, "<ResponseCode>50</ResponseCode>\r\n");
+      redir_stradd(dst, dstsize,
+		   "<ReplyMessage>Already logged on</ReplyMessage>\r\n");
+    }
+    redir_stradd(dst, dstsize, "</AuthenticationPollReply>\r\n");
+
+    if (conn->authenticated == 1) {
+      struct timeval timenow;
+      uint32_t sessiontime;
+      gettimeofday(&timenow, NULL);
+      sessiontime = timenow.tv_sec - conn->start_time.tv_sec;
+      sessiontime += (timenow.tv_usec - conn->start_time.tv_usec) / 1000000;
+      redir_stradd(dst, dstsize, "<SessionStatus>\r\n");
+      if (timeleft) {
+	redir_stradd(dst, dstsize, "<SessionTimeLeft>%d</SessionTimeLeft>\r\n",
+		     timeleft);
+      }
+      redir_stradd(dst, dstsize, "<Acct-Session-Time>%d</Acct-Session-Time>\r\n", sessiontime);
+      redir_stradd(dst, dstsize, "<Start-Time>%d</Start-Time>\r\n" , conn->start_time);
+      redir_stradd(dst, dstsize, "<Acct-Input-Octets>%d</Acct-Input-Octets>\r\n",
+		   conn->input_octets);
+      redir_stradd(dst, dstsize, "<Acct-Output-Octets>%d</Acct-Output-Octets>\r\n",
+		   conn->output_octets);
+      redir_stradd(dst, dstsize, "<Session-Timeout>%d</Session-Timeout>\r\n",
+		   conn->sessiontimeout);
+      redir_stradd(dst, dstsize, "<ChilliSpot-Max-Input-Octets>%d</ChilliSpot-Max-Input-Octets>\r\n",
+		   conn->maxinputoctets);
+
+      redir_stradd(dst, dstsize, "<ChilliSpot-Max-Output-Octets>%d</ChilliSpot-Max-Output-Octets>\r\n",
+		   conn->maxoutputoctets);
+      redir_stradd(dst, dstsize, "<ChilliSpot-Max-Total-Octets>%d</ChilliSpot-Max-Total-Octets>\r\n",
+		   conn->maxtotaloctets);
+      redir_stradd(dst, dstsize,
+		   "<LogoffURL>http://%s:%d/logoff</LogoffURL>\r\n",
+		   inet_ntoa(redir->addr), redir->port);
+      redir_stradd(dst, dstsize,
+		   "<StatusURL>http://%s:%d/status</StatusURL>\r\n",
+		   inet_ntoa(redir->addr), redir->port);
+      redir_stradd(dst, dstsize, "</SessionStatus>\r\n");
+    }
+    break;
   default:
     sys_err(LOG_ERR, __FILE__, __LINE__, 0, "Unknown res in switch");
     return -1;
@@ -334,10 +418,11 @@ static int redir_xmlreply(struct redir_t *redir, int res, char* hexchal,
 }
 
 /* Make an HTTP redirection reply and send it to the client */
-static int redir_reply(struct redir_t *redir, int fd, int res, 
-		       long int timeleft, char* hexchal, char* uid, 
-		       char* userurl, char* reply, char* redirurl,
-		       uint8_t *hismac) {
+static int redir_reply(struct redir_t *redir, int fd, 
+		       struct redir_conn_t *conn, int res, 
+		       long int timeleft,
+		       char* hexchal, char* uid, char* userurl, char* reply, 
+		       char* redirurl, uint8_t *hismac) {
   char buffer[REDIR_MAXBUFFER];
   char xmlreply[REDIR_MAXBUFFER];
   char *resp = NULL;
@@ -345,7 +430,7 @@ static int redir_reply(struct redir_t *redir, int fd, int res,
   buffer[0] = 0;
   xmlreply[0] = 0;
 
-  (void) redir_xmlreply(redir, res, hexchal, reply, redirurl,
+  (void) redir_xmlreply(redir, conn, res, timeleft, hexchal, reply, redirurl,
 			xmlreply, sizeof(xmlreply));
   
   switch (res) {
@@ -373,6 +458,12 @@ static int redir_reply(struct redir_t *redir, int fd, int res,
     break;
   case REDIR_ABOUT:
     break;
+  case REDIR_STATUS:
+    if (conn->authenticated == 1)
+      resp = "already";
+    else
+      resp = "notyet";
+    break;
   default:
     sys_err(LOG_ERR, __FILE__, __LINE__, 0, "Unknown res in switch");
     return -1;
@@ -575,6 +666,7 @@ void redir_set(struct redir_t *redir, int debug,
 	       struct in_addr *radiusserver0, struct in_addr *radiusserver1,
 	       uint16_t radiusauthport, uint16_t radiusacctport,
 	       char* radiussecret, char* radiusnasid,
+	       struct in_addr *radiusnasip, char* radiuscalled,
 	       char* radiuslocationid, char* radiuslocationname,
 	       int radiusnasporttype) {
 
@@ -590,6 +682,8 @@ void redir_set(struct redir_t *redir, int debug,
   redir->radiusacctport = radiusacctport;
   redir->radiussecret  = radiussecret;
   redir->radiusnasid  = radiusnasid;
+  redir->radiusnasip  = *radiusnasip;
+  redir->radiuscalled  = radiuscalled;
   redir->radiuslocationid  = radiuslocationid;
   redir->radiuslocationname  = radiuslocationname;
   redir->radiusnasporttype = radiusnasporttype;
@@ -721,6 +815,8 @@ static int redir_geturl(struct redir_t *redir, char *src, char *dst, int dstsize
   strncpy(dst+7+hostlen, path, pathlen);
   dst[7 + hostlen + pathlen] = 0;
 
+  if (optionsdebug) printf("Userurl: %s\n", dst);
+
   return 0;
 
 }
@@ -805,6 +901,7 @@ static int redir_getreq(struct redir_t *redir, int fd, struct redir_conn_t *conn
   char path[REDIR_MAXBUFFER];
   char resp[REDIR_MAXBUFFER];
   char mid[REDIR_MAXBUFFER];
+  int i;
 
   maxfd = fd;
 
@@ -841,6 +938,8 @@ static int redir_getreq(struct redir_t *redir, int fd, struct redir_conn_t *conn
       if (strstr(buffer, "\n")) break; /* Only interested in first line */
     }
   }
+  
+  for (i = 0; i<buflen; i++) if (buffer[i] == 0) buffer[i] = 0x0a; /* TODO: Hack to make Flash work */
 
   if (buflen <= 0) {
     if (optionsdebug) printf("No HTTP request received!\n");
@@ -851,6 +950,11 @@ static int redir_getreq(struct redir_t *redir, int fd, struct redir_conn_t *conn
     if (optionsdebug) printf("Could not parse path!\n");
     return -1;
   }
+
+  if (!redir_getparam(redir, buffer, "userurl",
+		     conn->userurl, sizeof(conn->userurl))) {
+    if (optionsdebug) printf("User URL: %s!\n", conn->userurl);
+  }
   
   if ((!strcmp(path, "logon")) || (!strcmp(path, "login"))) {
     if (redir_getparam(redir, buffer, "username", 
@@ -886,7 +990,7 @@ static int redir_getreq(struct redir_t *redir, int fd, struct redir_conn_t *conn
   else if (!strncmp(path, "msdownload", 10)) {
     conn->type = REDIR_MSDOWNLOAD;
     return 0;
-  } 
+  }
   else if (!strcmp(path, "prelogin")) {
     conn->type = REDIR_PRELOGIN;
     return 0;
@@ -899,6 +1003,10 @@ static int redir_getreq(struct redir_t *redir, int fd, struct redir_conn_t *conn
     conn->type = REDIR_ABOUT;
     return 0;
   }
+  else if (!strcmp(path, "status")) {
+    conn->type = REDIR_STATUS;
+    return 0;
+  }
   else {
     if (redir_geturl(redir, buffer, conn->userurl, sizeof(conn->userurl))) {
       if (optionsdebug) printf("Could not parse URL!\n");
@@ -967,6 +1075,16 @@ static int redir_cb_radius_auth_conf(struct radius_t *radius,
     return 0;
   }
 
+  /* Get Service Type */
+  if (!radius_getattr(pack, &attr, RADIUS_ATTR_SERVICE_TYPE, 0, 0, 0)) {
+    if(ntohl(attr->v.i) == RADIUS_SERVICE_TYPE_CHILLISPOT_AUTHORIZE_ONLY) {
+      sys_err(LOG_ERR, __FILE__, __LINE__, 0,
+	      "Chillispot-Authorize-Only Service-Type in Access-Accept");
+      conn->response = REDIR_FAILED_REJECT;
+      return 0;
+    }
+  }
+
   /* State */
   if (!radius_getattr(pack, &stateattr, RADIUS_ATTR_STATE, 0, 0, 0)) {
     conn->statelen = stateattr->l-2;
@@ -1003,7 +1121,21 @@ static int redir_cb_radius_auth_conf(struct radius_t *radius,
   else {
     conn->idletimeout = 0;
   }
-  
+
+  /* Filter ID */
+  if (!radius_getattr(pack, &attr, RADIUS_ATTR_FILTER_ID,
+		      0, 0, 0)) {
+    conn->filteridlen = attr->l-2;
+    memcpy(conn->filteridbuf, attr->v.t, attr->l-2);
+    conn->filteridbuf[attr->l-2] = 0;
+    conn->filterid = conn->filteridbuf;
+  }
+  else {
+    conn->filteridlen = 0;
+    conn->filteridbuf[0] = 0;
+    conn->filterid = NULL;
+  }
+
   /* Interim interval */
   if (!radius_getattr(pack, &interimattr, RADIUS_ATTR_ACCT_INTERIM_INTERVAL, 
 		      0, 0, 0)) {
@@ -1239,7 +1371,7 @@ static int redir_radius(struct redir_t *redir, struct in_addr *addr,
   }
 
   radius_addattr(radius, &radius_pack, RADIUS_ATTR_NAS_IP_ADDRESS, 0, 0,
-		 ntohl(conn->nasip.s_addr), NULL, 0); /* WISPr_V1.0 */
+		 ntohl(redir->radiusnasip.s_addr), NULL, 0); /* WISPr_V1.0 */
   radius_addattr(radius, &radius_pack, RADIUS_ATTR_SERVICE_TYPE, 0, 0,
 		 RADIUS_SERVICE_TYPE_LOGIN, NULL, 0); /* WISPr_V1.0 */
   radius_addattr(radius, &radius_pack, RADIUS_ATTR_FRAMED_IP_ADDRESS, 0, 0,
@@ -1254,14 +1386,9 @@ static int redir_radius(struct redir_t *redir, struct in_addr *addr,
   radius_addattr(radius, &radius_pack, RADIUS_ATTR_CALLING_STATION_ID, 0, 0, 0,
 		 (uint8_t*) mac, REDIR_MACSTRLEN);
   
-  /* Include our MAC address */
-  snprintf(mac, REDIR_MACSTRLEN+1, "%.2X-%.2X-%.2X-%.2X-%.2X-%.2X",
-	   conn->ourmac[0], conn->ourmac[1],
-	   conn->ourmac[2], conn->ourmac[3],
-	   conn->ourmac[4], conn->ourmac[5]);
-
-  radius_addattr(radius, &radius_pack, RADIUS_ATTR_CALLED_STATION_ID, 0, 0, 0,
-		 (uint8_t*) mac, REDIR_MACSTRLEN); /* WISPr_V1.0 */
+  if (redir->radiuscalled)
+    radius_addattr(radius, &radius_pack, RADIUS_ATTR_CALLED_STATION_ID, 0, 0, 0,
+		   (uint8_t*) redir->radiuscalled, strlen(redir->radiuscalled)); /* WISPr_V1.0 */
 
   if (redir->radiusnasid)
     radius_addattr(radius, &radius_pack, RADIUS_ATTR_NAS_IDENTIFIER, 0, 0, 0,
@@ -1493,7 +1620,8 @@ int redir_accept(struct redir_t *redir) {
     /* Was client was already logged on? */
     if (state == 1) {
       if (optionsdebug) printf("redir_accept: Already logged on\n");
-      redir_reply(redir, new_socket, REDIR_ALREADY, 0, NULL, NULL, NULL, NULL,
+      redir_reply(redir, new_socket, &conn, REDIR_ALREADY, 0, 
+		  NULL, NULL, conn.userurl, NULL,
 		  NULL, conn.hismac);
       redir_close();
       exit(0);
@@ -1510,8 +1638,9 @@ int redir_accept(struct redir_t *redir) {
 	exit(0);
       }
 
-      redir_reply(redir, new_socket, REDIR_FAILED_OTHER, 0, hexchal, NULL, 
-		  NULL, NULL, NULL, conn.hismac);
+      redir_reply(redir, new_socket, &conn, REDIR_FAILED_OTHER, 0,
+		  hexchal, NULL, conn.userurl, NULL,
+		  NULL, conn.hismac);
       redir_close();
       exit(0);
     }
@@ -1526,9 +1655,9 @@ int redir_accept(struct redir_t *redir) {
     if (optionsdebug) printf("Received radius reply\n");
 
     if (conn.response == REDIR_SUCCESS) { /* Radius-Accept */
-      redir_reply(redir, new_socket, conn.response, conn.sessiontimeout,
+      redir_reply(redir, new_socket, &conn, conn.response, conn.sessiontimeout,
                   NULL, conn.username, conn.userurl, conn.reply,
-                  conn.redirurl, conn.hismac); 
+                  conn.redirurl, conn.hismac);
       
       msg.type = REDIR_LOGIN;
       strncpy(msg.username, conn.username, sizeof(msg.username));
@@ -1547,6 +1676,8 @@ int redir_accept(struct redir_t *redir) {
       msg.maxoutputoctets = conn.maxoutputoctets;
       msg.maxtotaloctets = conn.maxtotaloctets;
       msg.sessionterminatetime = conn.sessionterminatetime;
+      msg.filteridlen = conn.filteridlen;
+      strncpy(msg.filteridbuf, conn.filteridbuf, sizeof(msg.filteridbuf));
       
       if (msgsnd(redir->msgid, (struct msgbuf*) &msg,
 		 sizeof(struct redir_msg_t), 0) < 0) {
@@ -1562,8 +1693,9 @@ int redir_accept(struct redir_t *redir) {
 	exit(0);
       }
 
-      redir_reply(redir, new_socket, conn.response, 0, hexchal, NULL, NULL, 
-		  conn.reply, NULL, conn.hismac);
+      redir_reply(redir, new_socket, &conn, conn.response, 0,
+		  hexchal, NULL, conn.userurl, conn.reply, 
+		  NULL, conn.hismac);
       redir_close();
       exit(0);
     }    
@@ -1579,7 +1711,8 @@ int redir_accept(struct redir_t *redir) {
       exit(0);
     }
 
-    redir_reply(redir, new_socket, REDIR_LOGOFF, 0, hexchal, NULL, NULL, NULL, 
+    redir_reply(redir, new_socket, &conn, REDIR_LOGOFF, 0, 
+		hexchal, NULL, conn.userurl, NULL, 
 		NULL, conn.hismac);
     redir_close();    
     exit(0); /* Terminate the client */
@@ -1593,20 +1726,23 @@ int redir_accept(struct redir_t *redir) {
     }
 
     if (state == 1) {
-      redir_reply(redir, new_socket, REDIR_ALREADY, 0, NULL, NULL, NULL, NULL,
+      redir_reply(redir, new_socket, &conn, REDIR_ALREADY, 0, 
+		  NULL, NULL, conn.userurl, NULL,
 		  NULL, conn.hismac);
     }
     else {
-      redir_reply(redir, new_socket, REDIR_NOTYET, 0, hexchal, NULL, NULL,
-		  NULL, NULL, conn.hismac);
+      redir_reply(redir, new_socket, &conn, REDIR_NOTYET, 0, 
+		  hexchal, NULL, conn.userurl, NULL,
+		  NULL, conn.hismac);
     }
     redir_close();
     exit(0);
   }
   else if (conn.type == REDIR_ABORT) {
     if (state == 1) {
-      redir_reply(redir, new_socket, REDIR_ABORT_NAK, 0, NULL, NULL, NULL, 
-		  NULL, NULL, conn.hismac);
+      redir_reply(redir, new_socket, &conn, REDIR_ABORT_NAK, 0,
+		  NULL, NULL, conn.userurl, NULL, 
+		  NULL, conn.hismac);
     }
     else {
       redir_memcopy(REDIR_ABORT);
@@ -1615,15 +1751,34 @@ int redir_accept(struct redir_t *redir) {
 	sys_err(LOG_ERR, __FILE__, __LINE__, errno, "msgsnd() failed!");
 	exit(0);
       }
-      redir_reply(redir, new_socket, REDIR_ABORT_ACK, 0, hexchal, NULL, NULL,
-		  NULL, NULL, conn.hismac);
+      redir_reply(redir, new_socket, &conn, REDIR_ABORT_ACK, 0, 
+		  hexchal, NULL, conn.userurl, NULL,
+		  NULL, conn.hismac);
     }
     redir_close();
     exit(0);
   }
   else if (conn.type == REDIR_ABOUT) {
-    redir_reply(redir, new_socket, REDIR_ABOUT, 0, NULL, NULL, NULL, 
-		NULL, NULL, NULL);
+    redir_reply(redir, new_socket, &conn, REDIR_ABOUT, 0, 
+		NULL, NULL, NULL, NULL,
+		NULL, NULL);
+    redir_close();
+    exit(0);
+  }
+  else if (conn.type == REDIR_STATUS) {
+    struct timeval timenow;
+    uint32_t sessiontime;
+    uint32_t timeleft;
+    gettimeofday(&timenow, NULL);
+    sessiontime = timenow.tv_sec - conn.start_time.tv_sec;
+    sessiontime += (timenow.tv_usec - conn.start_time.tv_usec) / 1000000;
+    if (conn.sessiontimeout)
+      timeleft = conn.sessiontimeout - sessiontime;
+    else
+      timeleft = 0;
+    redir_reply(redir, new_socket, &conn, REDIR_STATUS, timeleft,
+		NULL, NULL, NULL, NULL,
+		NULL, NULL);
     redir_close();
     exit(0);
   }
@@ -1678,12 +1833,14 @@ int redir_accept(struct redir_t *redir) {
     send(new_socket, buffer, buflen, 0);
   }
   else if (state == 1) {
-    redir_reply(redir, new_socket, REDIR_ALREADY, 0, NULL, NULL, NULL, NULL,
+    redir_reply(redir, new_socket, &conn, REDIR_ALREADY, 0,
+		NULL, NULL, conn.userurl, NULL,
 		NULL, conn.hismac);
   }
   else {
-    redir_reply(redir, new_socket, REDIR_NOTYET, 0, hexchal, NULL,
-		conn.userurl, NULL, NULL, conn.hismac);
+    redir_reply(redir, new_socket, &conn, REDIR_NOTYET, 0, 
+		hexchal, NULL, conn.userurl, NULL, 
+		NULL, conn.hismac);
   }
 
   redir_close();
